/*  Implementation of the Othello game
**
**  29/03/01, Christoph Zenger, Philippe Altherr
*/

with Predef
with Console
with Array
with String

var verbose: boolean := true

newtype OthelloPosition = {
   def put(x: int, y: int, m: int): ()
   def get(x: int, y: int): int
   def inc: ()
   def dec: ()
   def turn: int
   def changeTurn: ()
   def toString: String
   def allMoves: List[OthelloUndoMove]
   def elaborate(m: OthelloMove): OthelloUndoMove
   def negamax(p:OthelloPosition, a: int, b: int, d: int): int
   def score: int
   def eval: int
}

newtype OthelloMove = {
   def x: int
   def y: int
   def isValid: boolean
   def toString: String
}

newtype OthelloUndoMove = {
   def x: int
   def y: int
   def toString: String
   def exec(p: OthelloPosition): ()
   def undo(p: OthelloPosition): ()
   def isValid: boolean
}

type OthelloModule = {
   def OthelloBoard: OthelloPosition
   def stringMove(s: String): OthelloMove
   def OthelloMove(x: int, y:int): OthelloMove
   def OthelloPassMove: OthelloUndoMove
}

type UndoMoveList = List[OthelloUndoMove]
type TList = List[(int, int)]

let Othello:OthelloModule = {
   val red = 1
   val green = -1
   val empty = 0
   val OthelloPassMove: OthelloUndoMove = {
      val x = -1
      val y = -1
      val isValid = false
      def toString = "pass"
      val l = List.Nil
      def exec(p: OthelloPosition) = {
         p.changeTurn
      }
      def undo(p: OthelloPosition) = {
         p.changeTurn
      }
   }
   def OthelloBoard: OthelloPosition = {
      val pos = { 
         val board = create[Array[int]](8)
         for 0 7 1 {i|
            board(i) := createInitialized(8, empty)
         }
         board(3)(4) := red
         board(4)(3) := red
         board(3)(3) := green
         board(4)(4) := green
         board
      }
      var t:int := green
      var discs: int := 4
      def put(x: int, y: int, m: int) = {
         pos(x)(y) := m
      }
      def get(x: int, y: int): int = {
         pos(x)(y)
      }
      def inc = {
         discs := discs + 1
      }
      def dec = { 
         discs := discs - 1
      }
      def turn = t
      def changeTurn = {
         t := -t
      }
      def toString: String = {
         var s:String := "  a b c d e f g h\n"
         for 0 7 1 {i|{
            s := s + (8 - i) + " "
            for 0 7 1 {j|{
               val m = get(j,i)
               if (m == -1) {
                  s := s + "X "
               } else if (m == 1) {
                  s := s + "O "
               } else {
                  s := s + ". "
               }
            }}
            s := s + (8 - i) + "\n"
         }}
         s := s + "  a b c d e f g h\n"
         if ((turn) == green) {
            s := s + "X to move (? for help)\n"
         } else {
            s := s + "O to move (? for help)\n"
         }
         s := s + "The score is " + (score * turn) + ".\n"
         s := s + "The eval is " + (eval) + ".\n"
         s
      }


      def allMoves: UndoMoveList = {
        def allMovesi(i:int)(moves: UndoMoveList): UndoMoveList = {
          def allMovesj(j:int)(moves: UndoMoveList): UndoMoveList = {
            if (j > 7) moves else {
              val m = elaborate(OthelloMove(i,j))
              if (m.isValid) {
                allMovesj (j + 1) (m :: moves)
              } else {
                allMovesj (j + 1) (moves)
              }
            }
          }
          if (i > 7) moves else allMovesi(i + 1)(allMovesj 0 moves)
        }
        allMovesi 0 (List.Nil)
      }

      def score: int = {
        def score0(i:int)(s:int): int = {
          def score1(j:int)(s:int): int = {
            if (j > 7) s else score1(j+1)(s + get(i,j))
          }
          if (i > 7) s else score0(i+1)(score1 0 s)
        }
        score0 0 0
      }

      def elaborate(m: OthelloMove): OthelloUndoMove = {
         val x = m.x
         val y = m.y
         val l = {
            if (m.isValid) {
               if (get(m.x, m.y) == empty) {
                  def elab(dx:int, dy:int)(ts:TList):TList = {
                    def elab0(xs:TList)(u:int)(v:int):(TList, int, int) = {
                      if (inRange(u,v) && (get(u,v) == -(turn))) {
                        elab0 (List.Cons((u,v),xs)) (u + dx) (v + dy)
                      } else
                        (xs,u,v)
                    }
                    val u0 = (x + dx)
                    val v0 = (y + dy)
                    if (inRange(u0, v0) && (get(u0,v0) == -(turn))) {
                      val (xs,u,v) = elab0 (List.Nil[(int,int)]) u0 v0
                      if (inRange(u,v) && (get(u,v) == (turn))) {
                        ts ++ xs
                      } else
                        ts
                    } else
                      ts
                  }
                  elab(1,  1)(elab( 0,  1)(elab(-1,  1)(elab( 1,  0)
                    (elab(-1,  0)(elab( 1, -1)(elab( 0, -1)
                        (elab(-1, -1)(List.Nil[(int,int)]))))))))
               } else
                 List.Nil[(int,int)]
            } else
               List.Nil[(int,int)]
         }
         OthelloUndoMove(m, l)
      }

      def edge:int = {
        def edgei i:int b:int :int = {
          if (i > 7) b else {
            edgei (i + 1) (b + get(0,i) + get(i, 0) + get(i, 7) + get (7, i))
          }
        }
        edgei 0 0
      }

      def corner: int = {
         get(0,0) + get(7, 0) + get(0, 7) + get (7, 7)
      }
      def semicorner: int = {
         get(1,1) + get(6, 6) + get(1, 6) + get (6, 1)
      }
      def eval: int = {
         if (discs > 44) {
            turn * score
         } else {
            turn * (score + 3 * corner + 2 * edge - 4 * semicorner)
         }
      }
      def indent(n: int): () = {
         for 0 n 1 {i| print "  " }
      }
      def negamax(p: OthelloPosition, a: int, b: int, d: int): int = {
         if (d < 1) {
            eval
         } else {
            var all:List[OthelloUndoMove] := allMoves
            if (all.isEmpty) {
               OthelloPassMove.exec(p)
               val v = - negamax(p, b, a, d - 1)
               OthelloPassMove.undo(p)
               v
            } else {
               var max:int := a - 1
               while (| !all.isEmpty && max < b + 1) {|{
                  val mv = all.head
                  all := all.tail
                  mv.exec(p)
                  val v = - negamax(p, - b, - max - 1, d - 1)
                  mv.undo(p)
                  if (verbose) {
                     indent (4 - d)
                     print ("Move " + mv.toString + ": " + v)
                  } else {
                  }
                  if (v > max) {
                     max := v
                     if (verbose) {
                        println " (!)"
                     } else
                        ()
                  } else {
                     if (verbose) {
                        println ""
                     } else
                        ()
                  }
               }}
               max
            }
         }
      }
   }
   
   def OthelloMove(x1: int, y1:int): OthelloMove = {
      val x = x1
      val y = y1
      val isValid = inRange(x, y)
      def toString = {
         if (!isValid)
            "invalid move"
         else
            fromChar(charAt("abcdefgh")(x)) + (8 - y)
      }
   }
   def OthelloUndoMove(m:OthelloMove, l:List[(int, int)]):OthelloUndoMove = {
      val x = m.x
      val y = m.y
      val isValid = (m.isValid) && !(l.isEmpty)
      def exec(p: OthelloPosition): () = {
         p.put(x, y, p.turn)
         l.iterate {(u,v)|{
            p.put(u,v,p.turn)
         }}
         p.inc
         p.changeTurn
      }
      def undo(p: OthelloPosition): () = {
         p.put(x, y, empty)
         l.iterate {(u,v)|{
            p.put(u,v,p.turn)
         }}
         p.dec
         p.changeTurn
      }
      def toString: String = m.toString // + l.toString
   }
   def stringMove(s: String): OthelloMove = {
      if (length(s) == 2) {
         OthelloMove(charAt(s)(0) - 97, 7 - (charAt(s)(1) - 49))
      } else {
	 OthelloMove(-1, -1)
      }
   }
   def inRange(x: int, y:int):boolean =
         (-1 < x) && (x < 8) && (-1 < y) && (y < 8)
}

with Othello

var quit: boolean := false
var board: OthelloPosition := OthelloBoard

print(board.toString)
while (| !quit) {|{
   print "othello > "
   val s = readln
   if (length(s) > 0) {
      if (startsWith(s)("q")) {
         quit := true
      } else if (startsWith(s)("?")) {
         println "Enter "
         println "   \"b3\" to set a piece on b3,"
         println "   \"p\" to pass,"
         println "   \"m\" for a computer move (default search depth),"
         println "   \"m2\" for a computer move (search depth 2),"
         println "   \"v\" to toggle verbose,"
         println "   \"q\" to quit, and"
         println "   \"?\" for this help."
         println "Under Unix you can let othello play against itself with "
         println "   \"yes m2 | head -70 | funnel Othello\"."
      } else if (startsWith(s)("v")) {
         verbose := !verbose
	 if (verbose)
            println "verbose is now set to true."
         else
            println "verbose is now set to false."
      } else if (startsWith(s)("m")) {
         val depth = {
            if (length(s) > 1)
               charAt(s)(1) - '0'
            else
               3
         }
         val all = board.allMoves
         if (all.isEmpty) {
            OthelloPassMove.exec(board)
            print (board.toString)
            println "Computer passed."
         } else {
            var best: OthelloUndoMove := all.head
            var max: int := -100
            all.iterate {m|{
               m.exec(board)
               val v = - board.negamax(board, -100, - max - 1, depth - 1)
               print ("Move " + m.toString + ": " + v)
               if (v > max) {
                  max := v
                  best := m
                  println " (!)"
               } else
                  println ""
               m.undo(board)
            }}
            best.exec(board)
            print (board.toString)
            println ("Computer played " + best.toString + "(" + max + ";" + board.eval + ").")
         }
      } else if (startsWith(s)("p")) {
	 if (board.allMoves.isEmpty) {
            OthelloPassMove.exec(board)
         } else {
            println "This move is invalid in this position."
            println ("Choose from: " + (board.allMoves.map (m|m.toString)).toString)
         }
         print(board.toString)
      } else {
         val mv = stringMove(s)
         if (mv.isValid) {
            println(mv.toString)
            val emv = board.elaborate(mv)
            if (!emv.isValid) {
	       println "This move is invalid in this position."
               val all = board.allMoves
               if (all.isEmpty) {
                  println ("Choose from " + (all.map (m|m.toString)).toString + ".")
               } else {
                  println "Your only option is to pass. "
               }
            } else {
               emv.exec(board)
            }
            print(board.toString)
         } else {
	    println "This is an invalid command (or move)."
         }
      } 
   } else {
   }
}}

print(board.toString)

"Good Bye."
